@@ -16,7 +16,7 @@
 #include <dolfinx/graph/AdjacencyList.h>
 #include <dolfinx/mesh/Geometry.h>
 
-namespace
+namespace impl
 {
 
 /// Implementation of bc application (lifting) for an given a set of integration
@@ -28,8 +28,6 @@ namespace
 /// facets or interior facets in their specified format)
 /// @param[in] dofmap0 The dofmap for the rows of the matrix
 /// @param[in] dofmap0 The dofmap for the columns of the matrix
-/// @param[in] bs0 The block size for the rows
-/// @param[in] bs1 The block size for the columns
 /// @param[in] bc_values1 Array of Dirichlet condition values for dofs local to
 /// process
 /// @param[in] bc_markers1 Array indicating what dofs local to process is in a
@@ -41,32 +39,30 @@ namespace
 /// vector be, i.e. be <- be - scale * (A (g - x0))
 /// @tparam T Scalartype of local vector
 /// @tparam estride Stride in actiave entities
-template <typename T, std::size_t estride>
-void _lift_bc_entities(
+template <typename T, std::size_t estride, std::floating_point U>
+void lift_bc_entities(
     std::span<T> b, std::span<const std::int32_t> active_entities,
-    const dolfinx::graph::AdjacencyList<std::int32_t>& dofmap0,
-    const dolfinx::graph::AdjacencyList<std::int32_t>& dofmap1, int bs0,
-    int bs1, const std::span<const T>& bc_values1,
-    const std::vector<std::int8_t>& bc_markers1,
-    const std::shared_ptr<const dolfinx_mpc::MultiPointConstraint<T>>& mpc1,
+    const dolfinx::fem::DofMap& dofmap0, const dolfinx::fem::DofMap& dofmap1,
+    std::span<const T> bc_values1, std::span<const std::int8_t> bc_markers1,
+    const dolfinx_mpc::MultiPointConstraint<T, U>& mpc1,
     const std::function<const std::int32_t(std::span<const std::int32_t>)>
         fetch_cells,
     const std::function<void(std::span<T>, std::span<T>, const int, const int,
                              std::span<const std::int32_t>, std::size_t)>
         lift_local_vector)
 {
-
+  const int bs0 = dofmap0.bs();
+  const int bs1 = dofmap1.bs();
   // Get MPC data
   const std::shared_ptr<const dolfinx::graph::AdjacencyList<std::int32_t>>
-      masters = mpc1->masters();
+      masters = mpc1.masters();
   const std::shared_ptr<const dolfinx::graph::AdjacencyList<T>> coefficients
-      = mpc1->coefficients();
-  const std::vector<std::int8_t>& is_slave = mpc1->is_slave();
+      = mpc1.coefficients();
+  std::span<const std::int8_t> is_slave = mpc1.is_slave();
   const std::shared_ptr<const dolfinx::graph::AdjacencyList<std::int32_t>>
-      cell_to_slaves = mpc1->cell_to_slaves();
+      cell_to_slaves = mpc1.cell_to_slaves();
 
-  // NOTE: Assertion that all links have the same size (no P refinement)
-  const int num_dofs0 = dofmap0.links(0).size();
+  const int num_dofs0 = dofmap0.map().extent(1);
   std::vector<T> be;
   std::vector<T> be_copy;
   std::vector<T> Ae;
@@ -77,8 +73,8 @@ void _lift_bc_entities(
     auto entity = active_entities.subspan(e, estride);
     const std::int32_t cell = fetch_cells(entity);
     // Size data structure for assembly
-    auto dmap0 = dofmap0.links(cell);
-    auto dmap1 = dofmap1.links(cell);
+    auto dmap0 = dofmap0.cell_dofs(cell);
+    auto dmap1 = dofmap1.cell_dofs(cell);
     const int num_rows = bs0 * dmap0.size();
     const int num_cols = bs1 * dmap1.size();
     be.resize(num_rows);
@@ -86,19 +82,19 @@ void _lift_bc_entities(
 
     // Check if bc is applied to entity
     bool has_bc = false;
-    std::for_each(dmap1.begin(), dmap1.end(),
-                  [&bc_markers1, bs1, &has_bc](const auto dof)
-                  {
-                    for (int k = 0; k < bs1; ++k)
-                    {
-                      assert(bs1 * dof + k < (int)bc_markers1.size());
-                      if (bc_markers1[bs1 * dof + k])
-                      {
-                        has_bc = true;
-                        break;
-                      }
-                    }
-                  });
+    std::ranges::for_each(dmap1,
+                          [&bc_markers1, bs1, &has_bc](const auto dof)
+                          {
+                            for (int k = 0; k < bs1; ++k)
+                            {
+                              assert(bs1 * dof + k < (int)bc_markers1.size());
+                              if (bc_markers1[bs1 * dof + k])
+                              {
+                                has_bc = true;
+                                break;
+                              }
+                            }
+                          });
     if (!has_bc)
       continue;
 
@@ -107,14 +103,14 @@ void _lift_bc_entities(
     const std::span<T> _Ae(Ae);
     lift_local_vector(_be, _Ae, num_rows, num_cols, entity, e / estride);
     // Modify local element matrix if entity is connected to a slave cell
-    std::span<const int32_t> slaves = cell_to_slaves->links(cell);
+    std::span<const std::int32_t> slaves = cell_to_slaves->links(cell);
 
     if (slaves.size() > 0)
     {
       // Modify element vector for MPC and insert into b for non-local
       // contributions
       be_copy.resize(num_rows);
-      std::copy(be.begin(), be.end(), be_copy.begin());
+      std::ranges::copy(be, be_copy.begin());
       const std::span<T> _be_copy(be_copy);
       dolfinx_mpc::modify_mpc_vec<T>(b, _be, _be_copy, dmap0, dmap0.size(), bs0,
                                      is_slave, slaves, masters, coefficients);
@@ -140,18 +136,18 @@ void _lift_bc_entities(
 /// @param[in] x0 The function to subtract
 /// @param[in] scale Scale of lifting
 /// @param[in] mpc1 The multi point constraints
-template <typename T>
-void _apply_lifting(
+template <typename T, std::floating_point U>
+void apply_lifting(
     std::span<T> b, const std::shared_ptr<const dolfinx::fem::Form<T>> a,
     const std::vector<std::shared_ptr<const dolfinx::fem::DirichletBC<T>>>& bcs,
-    const std::span<const T>& x0, double scale,
-    const std::shared_ptr<const dolfinx_mpc::MultiPointConstraint<T>>& mpc1)
+    const std::span<const T>& x0, T scale,
+    const std::shared_ptr<const dolfinx_mpc::MultiPointConstraint<T, U>>& mpc1)
 {
   const std::vector<T> constants = pack_constants(*a);
   auto coeff_vec = dolfinx::fem::allocate_coefficient_storage(*a);
   dolfinx::fem::pack_coefficients(*a, coeff_vec);
   auto coefficients = dolfinx::fem::make_coefficients_span(coeff_vec);
-  const std::vector<std::int8_t>& is_slave = mpc1->is_slave();
+  std::span<const std::int8_t> is_slave = mpc1->is_slave();
 
   // Create 1D arrays of bc values and bc indicator
   std::vector<std::int8_t> bc_markers1;
@@ -167,19 +163,16 @@ void _apply_lifting(
   for (const std::shared_ptr<const dolfinx::fem::DirichletBC<T>>& bc : bcs)
   {
     bc->mark_dofs(bc_markers1);
-    bc->dof_values(bc_values1);
+    bc->set(bc_values1, std::nullopt, 1);
   }
 
   // Extract dofmaps for columns and rows of a
   assert(a->function_spaces().at(0));
-  const dolfinx::graph::AdjacencyList<std::int32_t>& dofmap1
-      = V1->dofmap()->list();
-  std::shared_ptr<const dolfinx::fem::FiniteElement> element1 = V1->element();
-  const dolfinx::graph::AdjacencyList<std::int32_t>& dofmap0
-      = a->function_spaces()[0]->dofmap()->list();
+  auto dofmap1 = V1->dofmap();
+  auto element1 = V1->element();
+  auto dofmap0 = a->function_spaces()[0]->dofmap();
   const int bs0 = a->function_spaces()[0]->dofmap()->bs();
-  std::shared_ptr<const dolfinx::fem::FiniteElement> element0
-      = a->function_spaces()[0]->element();
+  auto element0 = a->function_spaces()[0]->element();
 
   const bool needs_transformation_data
       = element0->needs_dof_transformations()
@@ -189,28 +182,35 @@ void _apply_lifting(
   auto mesh = a->function_spaces()[0]->mesh();
 
   // Prepare cell geometry
-  const dolfinx::graph::AdjacencyList<std::int32_t>& x_dofmap
-      = mesh->geometry().dofmap();
-  std::span<const double> x_g = mesh->geometry().x();
-  const int tdim = mesh->topology().dim();
+
+  MDSPAN_IMPL_STANDARD_NAMESPACE::mdspan<
+      const std::int32_t,
+      MDSPAN_IMPL_STANDARD_NAMESPACE::dextents<std::size_t, 2>>
+      x_dofmap = mesh->geometry().dofmap();
+  std::span<const U> x_g = mesh->geometry().x();
+  const int tdim = mesh->topology()->dim();
+  const std::size_t num_dofs_g = x_dofmap.extent(1);
+  std::vector<U> coordinate_dofs(3 * num_dofs_g);
 
   std::span<const std::uint32_t> cell_info;
   if (needs_transformation_data)
   {
-    mesh->topology_mutable().create_entity_permutations();
-    cell_info = std::span(mesh->topology().get_cell_permutation_info());
+    mesh->topology_mutable()->create_entity_permutations();
+    cell_info = std::span(mesh->topology()->get_cell_permutation_info());
   }
 
   // Get dof-transformations for the element matrix
   const std::function<void(const std::span<T>&,
                            const std::span<const std::uint32_t>&, std::int32_t,
                            int)>
-      dof_transform = element0->get_dof_transformation_function<T>();
+      dof_transform = element0->template dof_transformation_fn<T>(
+          dolfinx::fem::doftransform::standard);
   const std::function<void(const std::span<T>&,
                            const std::span<const std::uint32_t>&, std::int32_t,
                            int)>
       dof_transform_to_transpose
-      = element1->get_dof_transformation_to_transpose_function<T>();
+      = element1->template dof_transformation_right_fn<T>(
+          dolfinx::fem::doftransform::transpose);
 
   // Loop over cell integrals and lift bc
   if (a->num_integrals(dolfinx::fem::IntegralType::cell) > 0)
@@ -230,28 +230,25 @@ void _apply_lifting(
                 std::size_t index)
       {
         auto cell = entity.front();
-        // FIXME: Add proper interface for num coordinate dofs
-        const std::size_t num_dofs_g = x_dofmap.num_links(cell);
-        // FIXME: Reconsider when using mixed topology (mixed celltypes)
-        std::vector<double> coordinate_dofs(3 * num_dofs_g);
 
         // Fetch the coordinates of the cell
-        const std::span<const std::int32_t> x_dofs = x_dofmap.links(cell);
+        auto x_dofs = MDSPAN_IMPL_STANDARD_NAMESPACE::submdspan(
+            x_dofmap, cell, MDSPAN_IMPL_STANDARD_NAMESPACE::full_extent);
         for (std::size_t i = 0; i < x_dofs.size(); ++i)
         {
-          std::copy_n(
-              std::next(x_g.begin(), 3 * x_dofs[i]),3,
-              std::next(coordinate_dofs.begin(), 3 * i));
+          std::ranges::copy_n(std::next(x_g.begin(), 3 * x_dofs[i]), 3,
+                              std::next(coordinate_dofs.begin(), 3 * i));
         }
+
         // Tabulate tensor
-        std::fill(Ae.data(), Ae.data() + Ae.size(), 0);
+        std::ranges::fill(Ae, 0);
         kernel(Ae.data(), coeffs.first.data() + index * coeffs.second,
                constants.data(), coordinate_dofs.data(), nullptr, nullptr);
         dof_transform(Ae, cell_info, cell, num_cols);
         dof_transform_to_transpose(Ae, cell_info, cell, num_rows);
 
-        auto dmap1 = dofmap1.links(cell);
-        std::fill(be.begin(), be.end(), 0);
+        auto dmap1 = dofmap1->cell_dofs(cell);
+        std::ranges::fill(be, 0);
         for (std::size_t j = 0; j < dmap1.size(); ++j)
         {
           for (std::int32_t k = 0; k < bs1; k++)
@@ -279,9 +276,12 @@ void _apply_lifting(
         }
       };
       // Assemble over all active cells
-      const std::vector<std::int32_t>& cells = a->cell_domains(i);
-      _lift_bc_entities<T, 1>(b, cells, dofmap0, dofmap1, bs0, bs1, bc_values1,
-                              bc_markers1, mpc1, fetch_cells, lift_bcs_cell);
+      std::span<const std::int32_t> cells
+          = a->domain(dolfinx::fem::IntegralType::cell, i);
+      lift_bc_entities<T, 1>(b, cells, *dofmap0, *dofmap1,
+                             std::span<const T>(bc_values1),
+                             std::span<const std::int8_t>(bc_markers1), *mpc1,
+                             fetch_cells, lift_bcs_cell);
     }
   }
 
@@ -293,9 +293,7 @@ void _apply_lifting(
         = [&](std::span<const std::int32_t> entity) { return entity.front(); };
 
     // Get number of cells per facet to be able to get the facet permutation
-    const int tdim = mesh->topology().dim();
-    const int num_cell_facets = dolfinx::mesh::cell_num_entities(
-        mesh->topology().cell_type(), tdim - 1);
+    const int tdim = mesh->topology()->dim();
     for (int i : a->integral_ids(dolfinx::fem::IntegralType::exterior_facet))
     {
       const auto& coeffs
@@ -316,28 +314,25 @@ void _apply_lifting(
         // Fetch the coordiantes of the cell
         const std::int32_t cell = entity[0];
         const int local_facet = entity[1];
-        const std::span<const std::int32_t> x_dofs = x_dofmap.links(cell);
-        // FIXME: Add proper interface for num coordinate dofs
-        const std::size_t num_dofs_g = x_dofmap.num_links(cell);
-        // FIXME: Reconsider when using mixed topology (mixed celltypes)
 
-        std::vector<double> coordinate_dofs(3 * num_dofs_g);
+        // Fetch the coordinates of the cell
+        auto x_dofs = MDSPAN_IMPL_STANDARD_NAMESPACE::submdspan(
+            x_dofmap, cell, MDSPAN_IMPL_STANDARD_NAMESPACE::full_extent);
         for (std::size_t i = 0; i < x_dofs.size(); ++i)
         {
-          std::copy_n(
-              std::next(x_g.begin(), 3 * x_dofs[i]),3,
-              std::next(coordinate_dofs.begin(), 3 * i));
+          std::ranges::copy_n(std::next(x_g.begin(), 3 * x_dofs[i]), 3,
+                              std::next(coordinate_dofs.begin(), 3 * i));
         }
 
         // Tabulate tensor
-        std::fill(Ae.data(), Ae.data() + Ae.size(), 0);
+        std::ranges::fill(Ae, 0);
         kernel(Ae.data(), coeffs.first.data() + index * coeffs.second,
                constants.data(), coordinate_dofs.data(), &local_facet, nullptr);
         dof_transform(Ae, cell_info, cell, num_cols);
         dof_transform_to_transpose(Ae, cell_info, cell, num_rows);
 
-        auto dmap1 = dofmap1.links(cell);
-        std::fill(be.begin(), be.end(), 0);
+        auto dmap1 = dofmap1->cell_dofs(cell);
+        std::ranges::fill(be, 0);
         for (std::size_t j = 0; j < dmap1.size(); ++j)
         {
           for (std::int32_t k = 0; k < bs1; k++)
@@ -367,11 +362,11 @@ void _apply_lifting(
       };
 
       // Assemble over all active cells
-      const std::vector<std::int32_t>& active_facets
-          = a->exterior_facet_domains(i);
-      _lift_bc_entities<T, 2>(b, active_facets, dofmap0, dofmap1, bs0, bs1,
-                              bc_values1, bc_markers1, mpc1, fetch_cell,
-                              lift_bc_exterior_facet);
+      std::span<const std::int32_t> active_facets
+          = a->domain(dolfinx::fem::IntegralType::exterior_facet, i);
+      impl::lift_bc_entities<T, 2>(b, active_facets, *dofmap0, *dofmap1,
+                                   bc_values1, bc_markers1, *mpc1, fetch_cell,
+                                   lift_bc_exterior_facet);
     }
   }
   if (a->num_integrals(dolfinx::fem::IntegralType::interior_facet) > 0)
@@ -385,14 +380,14 @@ void _apply_lifting(
     //     mesh->topology_mutable().create_connectivity(tdim - 1, tdim);
     //     mesh->topology_mutable().create_entity_permutations();
     //     const std::vector<std::uint8_t>& perms
-    //         = mesh->topology().get_facet_permutations();
+    //         = mesh->topology()->get_facet_permutations();
     //     get_perm = [&perms](std::size_t i) { return perms[i]; };
     //   }
     //   else
     //     get_perm = [](std::size_t) { return 0; };
   }
 }
-} // namespace
+} // namespace impl
 
 namespace dolfinx_mpc
 {
@@ -423,7 +418,8 @@ void apply_lifting(
         std::vector<std::shared_ptr<const dolfinx::fem::DirichletBC<double>>>>&
         bcs1,
     const std::vector<std::span<const double>>& x0, double scale,
-    const std::shared_ptr<const dolfinx_mpc::MultiPointConstraint<double>>& mpc)
+    const std::shared_ptr<
+        const dolfinx_mpc::MultiPointConstraint<double, double>>& mpc)
 {
   if (!x0.empty() and x0.size() != a.size())
   {
@@ -440,12 +436,12 @@ void apply_lifting(
   {
     if (x0.empty())
     {
-      _apply_lifting<double>(b, a[j], bcs1[j], std::span<const double>(), scale,
-                             mpc);
+      impl::apply_lifting<double>(b, a[j], bcs1[j], std::span<const double>(),
+                                  scale, mpc);
     }
     else
     {
-      _apply_lifting<double>(b, a[j], bcs1[j], x0[j], scale, mpc);
+      impl::apply_lifting<double>(b, a[j], bcs1[j], x0[j], scale, mpc);
     }
   }
 }
@@ -475,10 +471,12 @@ void apply_lifting(
         a,
     const std::vector<std::vector<std::shared_ptr<
         const dolfinx::fem::DirichletBC<std::complex<double>>>>>& bcs1,
-    const std::vector<std::span<const std::complex<double>>>& x0, double scale,
+    const std::vector<std::span<const std::complex<double>>>& x0,
+    std::complex<double> scale,
 
     const std::shared_ptr<
-        const dolfinx_mpc::MultiPointConstraint<std::complex<double>>>& mpc)
+        const dolfinx_mpc::MultiPointConstraint<std::complex<double>, double>>&
+        mpc)
 {
   if (!x0.empty() and x0.size() != a.size())
   {
@@ -495,14 +493,129 @@ void apply_lifting(
   {
     if (x0.empty())
     {
-      _apply_lifting<std::complex<double>>(
+      impl::apply_lifting<std::complex<double>>(
           b, a[j], bcs1[j], std::span<const std::complex<double>>(), scale,
           mpc);
     }
     else
     {
-      _apply_lifting<std::complex<double>>(b, a[j], bcs1[j], x0[j], scale, mpc);
+      impl::apply_lifting<std::complex<double>>(b, a[j], bcs1[j], x0[j], scale,
+                                                mpc);
     }
   }
 }
+
+/// Modify b such that:
+///
+///   b <- b - scale * K^T (A_j (g_j 0 x0_j))
+///
+/// where j is a block (nest) row index and K^T is the reduction matrix stemming
+/// from the multi point constraint. For non - blocked problems j = 0.
+/// The boundary conditions bcs1 are on the trial spaces V_j.
+/// The forms in [a] must have the same test space as L (from
+/// which b was built), but the trial space may differ. If x0 is not
+/// supplied, then it is treated as 0.
+/// @param[in,out] b The vector to be modified
+/// @param[in] a The bilinear formss, where a[j] is the form that
+/// generates A[j]
+/// @param[in] bcs List of boundary conditions for each block, i.e. bcs1[2]
+/// are the boundary conditions applied to the columns of a[2] / x0[2]
+/// block.
+/// @param[in] x0 The vectors used in the lifitng.
+/// @param[in] scale Scaling to apply
+/// @param[in] mpc The multi point constraints
+void apply_lifting(
+    std::span<float> b,
+    const std::vector<std::shared_ptr<const dolfinx::fem::Form<float>>> a,
+    const std::vector<
+        std::vector<std::shared_ptr<const dolfinx::fem::DirichletBC<float>>>>&
+        bcs1,
+    const std::vector<std::span<const float>>& x0, float scale,
+    const std::shared_ptr<
+        const dolfinx_mpc::MultiPointConstraint<float, float>>& mpc)
+{
+  if (!x0.empty() and x0.size() != a.size())
+  {
+    throw std::runtime_error(
+        "Mismatch in size between x0 and bilinear form in assembler.");
+  }
+
+  if (a.size() != bcs1.size())
+  {
+    throw std::runtime_error(
+        "Mismatch in size between a and bcs in assembler.");
+  }
+  for (std::size_t j = 0; j < a.size(); ++j)
+  {
+    if (x0.empty())
+    {
+      impl::apply_lifting<float>(b, a[j], bcs1[j], std::span<const float>(),
+                                 scale, mpc);
+    }
+    else
+    {
+      impl::apply_lifting<float>(b, a[j], bcs1[j], x0[j], scale, mpc);
+    }
+  }
+}
+/// Modify b such that:
+///
+///   b <- b - scale * K^T (A_j (g_j 0 x0_j))
+///
+/// where j is a block (nest) row index and K^T is the reduction matrix stemming
+/// from the multi point constraint. For non - blocked problems j = 0.
+/// The boundary conditions bcs1 are on the trial spaces V_j.
+/// The forms in [a] must have the same test space as L (from
+/// which b was built), but the trial space may differ. If x0 is not
+/// supplied, then it is treated as 0.
+/// @param[in,out] b The vector to be modified
+/// @param[in] a The bilinear formss, where a[j] is the form that
+/// generates A[j]
+/// @param[in] bcs List of boundary conditions for each block, i.e. bcs1[2]
+/// are the boundary conditions applied to the columns of a[2] / x0[2]
+/// block.
+/// @param[in] x0 The vectors used in the lifitng.
+/// @param[in] scale Scaling to apply
+/// @param[in] mpc The multi point constraints
+void apply_lifting(
+    std::span<std::complex<float>> b,
+    const std::vector<
+        std::shared_ptr<const dolfinx::fem::Form<std::complex<float>>>>
+        a,
+    const std::vector<std::vector<
+        std::shared_ptr<const dolfinx::fem::DirichletBC<std::complex<float>>>>>&
+        bcs1,
+    const std::vector<std::span<const std::complex<float>>>& x0,
+    std::complex<float> scale,
+
+    const std::shared_ptr<
+        const dolfinx_mpc::MultiPointConstraint<std::complex<float>, float>>&
+        mpc)
+{
+  if (!x0.empty() and x0.size() != a.size())
+  {
+    throw std::runtime_error(
+        "Mismatch in size between x0 and bilinear form in assembler.");
+  }
+
+  if (a.size() != bcs1.size())
+  {
+    throw std::runtime_error(
+        "Mismatch in size between a and bcs in assembler.");
+  }
+  for (std::size_t j = 0; j < a.size(); ++j)
+  {
+    if (x0.empty())
+    {
+      impl::apply_lifting<std::complex<float>>(
+          b, a[j], bcs1[j], std::span<const std::complex<float>>(), scale, mpc);
+    }
+    else
+    {
+      impl::apply_lifting<std::complex<float>>(b, a[j], bcs1[j], x0[j], scale,
+                                               mpc);
+    }
+  }
+}
+
 } // namespace dolfinx_mpc
\ No newline at end of file
