@@ -3,24 +3,28 @@
 # This file is part of DOLFINX_MPC
 #
 # SPDX-License-Identifier:    MIT
+from __future__ import annotations
 
-from typing import Optional, Sequence, Union
+from collections.abc import Sequence
+from typing import Optional, Union
+
+from petsc4py import PETSc as _PETSc
 
 import dolfinx.cpp as _cpp
 import dolfinx.fem as _fem
-from petsc4py import PETSc as _PETSc
 
 from dolfinx_mpc import cpp
 
 from .multipointconstraint import MultiPointConstraint
 
 
-def assemble_matrix(form: _fem.FormMetaClass,
-                    constraint: Union[MultiPointConstraint,
-                                      Sequence[MultiPointConstraint]],
-                    bcs: Sequence[_fem.DirichletBCMetaClass] = [],
-                    diagval: _PETSc.ScalarType = 1,
-                    A: Optional[_PETSc.Mat] = None) -> _PETSc.Mat:
+def assemble_matrix(
+    form: _fem.Form,
+    constraint: Union[MultiPointConstraint, Sequence[MultiPointConstraint]],
+    bcs: Optional[Sequence[_fem.DirichletBC]] = None,
+    diagval: _PETSc.ScalarType = 1,  # type: ignore
+    A: Optional[_PETSc.Mat] = None,  # type: ignore
+) -> _PETSc.Mat:  # type: ignore
     """
     Assemble a compiled DOLFINx bilinear form into a PETSc matrix with corresponding multi point constraints
     and Dirichlet boundary conditions.
@@ -33,35 +37,32 @@ def assemble_matrix(form: _fem.FormMetaClass,
         A: PETSc matrix to assemble into
 
     Returns:
-        _PETSc.Mat: The matrix with the assembled bi-linear form
+        _PETSc.Mat: The matrix with the assembled bi-linear form  #type: ignore
     """
+    bcs = [] if bcs is None else [bc._cpp_object for bc in bcs]
     if not isinstance(constraint, Sequence):
         assert form.function_spaces[0] == form.function_spaces[1]
         constraint = (constraint, constraint)
 
     # Generate matrix with MPC sparsity pattern
     if A is None:
-        A = cpp.mpc.create_matrix(form, constraint[0]._cpp_object,
-                                  constraint[1]._cpp_object)
+        A = cpp.mpc.create_matrix(form._cpp_object, constraint[0]._cpp_object, constraint[1]._cpp_object)
     A.zeroEntries()
 
     # Assemble matrix in C++
-    cpp.mpc.assemble_matrix(A, form, constraint[0]._cpp_object,
-                            constraint[1]._cpp_object, bcs, diagval)
+    cpp.mpc.assemble_matrix(A, form._cpp_object, constraint[0]._cpp_object, constraint[1]._cpp_object, bcs, diagval)
 
     # Add one on diagonal for Dirichlet boundary conditions
     if form.function_spaces[0] is form.function_spaces[1]:
-        A.assemblyBegin(_PETSc.Mat.AssemblyType.FLUSH)
-        A.assemblyEnd(_PETSc.Mat.AssemblyType.FLUSH)
+        A.assemblyBegin(_PETSc.Mat.AssemblyType.FLUSH)  # type: ignore
+        A.assemblyEnd(_PETSc.Mat.AssemblyType.FLUSH)  # type: ignore
         _cpp.fem.petsc.insert_diagonal(A, form.function_spaces[0], bcs, diagval)
 
     A.assemble()
     return A
 
 
-def create_sparsity_pattern(form: _fem.FormMetaClass,
-                            mpc: Union[MultiPointConstraint,
-                                       Sequence[MultiPointConstraint]]):
+def create_sparsity_pattern(form: _fem.Form, mpc: Union[MultiPointConstraint, Sequence[MultiPointConstraint]]):
     """
     Create sparsity-pattern for MPC given a compiled DOLFINx form
 
@@ -70,23 +71,21 @@ def create_sparsity_pattern(form: _fem.FormMetaClass,
         mpc: For square forms, the MPC. For rectangular forms a list of 2 MPCs on
             axis 0 & 1, respectively
     """
-    if isinstance(mpc, MultiPointConstraint):
-        mpc._not_finalized()
-        return cpp.mpc.create_sparsity_pattern(form, mpc._cpp_object,
-                                               mpc._cpp_object)
-    elif isinstance(mpc, list):
+    if isinstance(mpc, Sequence):
         assert len(mpc) == 2
         for mpc_ in mpc:
-            mpc_._not_finalized()
-            return cpp.mpc.create_sparsity_pattern(form, mpc[0]._cpp_object,
-                                                   mpc[1]._cpp_object)
+            mpc_._not_finalized()  # type: ignore
+            return cpp.mpc.create_sparsity_pattern(form._cpp_object, mpc[0]._cpp_object, mpc[1]._cpp_object)
     else:
-        raise ValueError(f"Unknown input type {type(mpc)}")
+        mpc._not_finalized()  # type: ignore
+        return cpp.mpc.create_sparsity_pattern(
+            form._cpp_object,
+            mpc._cpp_object,  # type: ignore
+            mpc._cpp_object,  # type: ignore
+        )  # type: ignore
 
 
-def create_matrix_nest(
-        a: Sequence[Sequence[_fem.FormMetaClass]],
-        constraints: Sequence[MultiPointConstraint]):
+def create_matrix_nest(a: Sequence[Sequence[_fem.Form]], constraints: Sequence[MultiPointConstraint]):
     """
     Create a PETSc matrix of type "nest" with appropriate sparsity pattern
     given the provided multi points constraints
@@ -104,19 +103,22 @@ def create_matrix_nest(
             if a[i][j] is None:
                 continue
             A_[i][j] = cpp.mpc.create_matrix(
-                a[i][j], constraints[i]._cpp_object, constraints[j]._cpp_object)
+                a[i][j]._cpp_object, constraints[i]._cpp_object, constraints[j]._cpp_object
+            )
 
-    A = _PETSc.Mat().createNest(
-        A_, comm=constraints[0].function_space.mesh.comm)
+    A = _PETSc.Mat().createNest(  # type: ignore
+        A_, comm=constraints[0].function_space.mesh.comm
+    )
     return A
 
 
 def assemble_matrix_nest(
-        A: _PETSc.Mat,
-        a: Sequence[Sequence[_fem.FormMetaClass]],
-        constraints: Sequence[MultiPointConstraint],
-        bcs: Sequence[_fem.DirichletBCMetaClass] = [],
-        diagval: _PETSc.ScalarType = 1):
+    A: _PETSc.Mat,  # type: ignore
+    a: Sequence[Sequence[_fem.Form]],
+    constraints: Sequence[MultiPointConstraint],
+    bcs: Sequence[_fem.DirichletBC] = [],
+    diagval: _PETSc.ScalarType = 1,  # type: ignore
+):
     """
     Assemble a compiled DOLFINx bilinear form into a PETSc matrix of type
     "nest" with corresponding multi point constraints and Dirichlet boundary
@@ -133,6 +135,4 @@ def assemble_matrix_nest(
         for j, a_block in enumerate(a_row):
             if a_block is not None:
                 Asub = A.getNestSubMatrix(i, j)
-                assemble_matrix(
-                    a_block, (constraints[i], constraints[j]),
-                    bcs=bcs, diagval=diagval, A=Asub)
+                assemble_matrix(a_block, (constraints[i], constraints[j]), bcs=bcs, diagval=diagval, A=Asub)
