@@ -3,23 +3,34 @@
 # This file is part of DOLFINX_MPC
 #
 # SPDX-License-Identifier:    MIT
+from __future__ import annotations
 
 import contextlib
 from typing import List, Optional, Sequence
 
+from petsc4py import PETSc as _PETSc
+
 import dolfinx.cpp as _cpp
 import dolfinx.fem as _fem
+import dolfinx.la as _la
+import numpy
 import ufl
+from dolfinx import default_scalar_type
 from dolfinx.common import Timer
-from petsc4py import PETSc as _PETSc
 
 import dolfinx_mpc.cpp
 
-from .multipointconstraint import MultiPointConstraint
+from .multipointconstraint import MultiPointConstraint, _float_classes
 
 
-def apply_lifting(b: _PETSc.Vec, form: List[_fem.FormMetaClass], bcs: List[List[_fem.DirichletBCMetaClass]],
-                  constraint: MultiPointConstraint, x0: List[_PETSc.Vec] = [], scale: float = 1.0):
+def apply_lifting(
+    b: _PETSc.Vec,  # type: ignore
+    form: List[_fem.Form],
+    bcs: List[List[_fem.DirichletBC]],
+    constraint: MultiPointConstraint,
+    x0: List[_PETSc.Vec] = [],  # type: ignore
+    scale: _float_classes = default_scalar_type(1.0),
+):  # type: ignore
     """
     Apply lifting to vector b, i.e.
     :math:`b = b - scale \\cdot K^T (A_j (g_j - x0_j))`
@@ -32,18 +43,24 @@ def apply_lifting(b: _PETSc.Vec, form: List[_fem.FormMetaClass], bcs: List[List[
         x0: List of vectors
         scale: Scaling for lifting
     """
+    if isinstance(scale, numpy.generic):  # nanobind conversion of numpy dtypes to general Python types
+        scale = scale.item()  # type: ignore
     t = Timer("~MPC: Apply lifting (C++)")
     with contextlib.ExitStack() as stack:
         x0 = [stack.enter_context(x.localForm()) for x in x0]
         x0_r = [x.array_r for x in x0]
         b_local = stack.enter_context(b.localForm())
-        dolfinx_mpc.cpp.mpc.apply_lifting(b_local.array_w, form,
-                                          bcs, x0_r, scale, constraint._cpp_object)
+        _forms = [f._cpp_object for f in form]
+        _bcs = [[bc._cpp_object for bc in bcs0] for bcs0 in bcs]
+        dolfinx_mpc.cpp.mpc.apply_lifting(b_local.array_w, _forms, _bcs, x0_r, scale, constraint._cpp_object)
     t.stop()
 
 
-def assemble_vector(form: ufl.form.Form, constraint: MultiPointConstraint,
-                    b: Optional[_PETSc.Vec] = None) -> _PETSc.Vec:
+def assemble_vector(
+    form: ufl.form.Form,
+    constraint: MultiPointConstraint,
+    b: Optional[_PETSc.Vec] = None,  # type: ignore
+) -> _PETSc.Vec:  # type: ignore
     """
     Assemble a linear form into vector `b` with corresponding multi point constraint
 
@@ -57,19 +74,19 @@ def assemble_vector(form: ufl.form.Form, constraint: MultiPointConstraint,
     """
 
     if b is None:
-        b = _cpp.la.petsc.create_vector(constraint.function_space.dofmap.index_map,
-                                        constraint.function_space.dofmap.index_map_bs)
+        b = _la.create_petsc_vector(
+            constraint.function_space.dofmap.index_map,
+            constraint.function_space.dofmap.index_map_bs,
+        )
     t = Timer("~MPC: Assemble vector (C++)")
     with b.localForm() as b_local:
         b_local.set(0.0)
-        dolfinx_mpc.cpp.mpc.assemble_vector(b_local, form, constraint._cpp_object)
+        dolfinx_mpc.cpp.mpc.assemble_vector(b_local.array_w, form._cpp_object, constraint._cpp_object)
     t.stop()
     return b
 
 
-def create_vector_nest(
-        L: Sequence[_fem.FormMetaClass],
-        constraints: Sequence[MultiPointConstraint]) -> _PETSc.Vec:
+def create_vector_nest(L: Sequence[_fem.Form], constraints: Sequence[MultiPointConstraint]) -> _PETSc.Vec:  # type: ignore
     """
     Create a PETSc vector of type "nest" appropriate for the provided multi
     point constraints
@@ -79,20 +96,22 @@ def create_vector_nest(
         constraints: An ordered list of multi point constraints
 
     Returns:
-        PETSc.Vec: A PETSc vector of type "nest"
+        PETSc.Vec: A PETSc vector of type "nest"  #type: ignore
     """
     assert len(constraints) == len(L)
 
-    maps = [(constraint.function_space.dofmap.index_map,
-             constraint.function_space.dofmap.index_map_bs)
-            for constraint in constraints]
+    maps = [
+        (constraint.function_space.dofmap.index_map, constraint.function_space.dofmap.index_map_bs)
+        for constraint in constraints
+    ]
     return _cpp.fem.petsc.create_vector_nest(maps)
 
 
 def assemble_vector_nest(
-        b: _PETSc.Vec,
-        L: Sequence[_fem.FormMetaClass],
-        constraints: Sequence[MultiPointConstraint]):
+    b: _PETSc.Vec,  # type: ignore
+    L: Sequence[_fem.Form],
+    constraints: Sequence[MultiPointConstraint],
+):
     """
     Assemble a linear form into a PETSc vector of type "nest"
 
