@@ -6,21 +6,25 @@
 #
 # Multi point constraint problem for linear elasticity with slip conditions
 # between two cubes.
+from __future__ import annotations
 
-import dolfinx.fem as fem
-import dolfinx_mpc.utils
-import gmsh
-import numpy as np
-import pytest
-import scipy.sparse.linalg
-import ufl
-from dolfinx.common import Timer, TimingType, list_timings
-from dolfinx.io import gmshio
-from dolfinx_mpc.utils import get_assemblers  # noqa: F401
 from mpi4py import MPI
 from petsc4py import PETSc
 
+import dolfinx.fem as fem
+import gmsh
+import numpy as np
+import numpy.testing as nt
+import pytest
+import scipy.sparse.linalg
+import ufl
+from dolfinx import default_scalar_type
+from dolfinx.common import Timer, TimingType, list_timings
+from dolfinx.io import gmshio
+
 import dolfinx_mpc
+import dolfinx_mpc.utils
+from dolfinx_mpc.utils import get_assemblers  # noqa: F401
 
 theta = np.pi / 5
 
@@ -52,17 +56,14 @@ def generate_hex_boxes():
         top = gmsh.model.occ.addRectangle(x0, y0, z2, x1 - x0, y1 - y0)
 
         # Set mesh size at point
-        gmsh.model.occ.extrude([(2, bottom)], 0, 0, z1 - z0,
-                               numElements=[int(1 / (2 * res))], recombine=True)
-        gmsh.model.occ.extrude([(2, top)], 0, 0, z1 - z2 - 1e-12,
-                               numElements=[int(1 / (2 * res))], recombine=True)
+        gmsh.model.occ.extrude([(2, bottom)], 0, 0, z1 - z0, numElements=[int(1 / (2 * res))], recombine=True)
+        gmsh.model.occ.extrude([(2, top)], 0, 0, z1 - z2 - 1e-12, numElements=[int(1 / (2 * res))], recombine=True)
         # Syncronize to be able to fetch entities
         gmsh.model.occ.synchronize()
 
         # Create entity -> marker map (to be used after rotation)
         volumes = gmsh.model.getEntities(3)
-        volume_entities = {"Top": [None, volume_markers[1]],
-                           "Bottom": [None, volume_markers[0]]}
+        volume_entities = {"Top": [None, volume_markers[1]], "Bottom": [None, volume_markers[0]]}
         for i, volume in enumerate(volumes):
             com = gmsh.model.occ.getCenterOfMass(volume[0], volume[1])
             if np.isclose(com[2], (z1 - z0) / 2):
@@ -72,13 +73,14 @@ def generate_hex_boxes():
                 top_index = i
                 volume_entities["Top"][0] = volume
         surfaces = ["Top", "Bottom", "Left", "Right", "Front", "Back"]
-        entities = {"Bottom": {key: [[], None] for key in surfaces},
-                    "Top": {key: [[], None] for key in surfaces}}
+        entities = {
+            "Bottom": {key: [[], None] for key in surfaces},
+            "Top": {key: [[], None] for key in surfaces},
+        }
         # Identitfy entities for each surface of top and bottom cube
 
         # Physical markers for bottom cube
-        bottom_surfaces = gmsh.model.getBoundary([volumes[bottom_index]],
-                                                 recursive=False, oriented=False)
+        bottom_surfaces = gmsh.model.getBoundary([volumes[bottom_index]], recursive=False, oriented=False)
         for entity in bottom_surfaces:
             com = gmsh.model.occ.getCenterOfMass(entity[0], entity[1])
             if np.allclose(com, [(x1 - x0) / 2, y1, (z1 - z0) / 2]):
@@ -100,8 +102,7 @@ def generate_hex_boxes():
                 entities["Bottom"]["Front"][0].append(entity[1])
                 entities["Bottom"]["Front"][1] = facet_markers[0][5]
         # Physical markers for top
-        top_surfaces = gmsh.model.getBoundary([volumes[top_index]],
-                                              recursive=False, oriented=False)
+        top_surfaces = gmsh.model.getBoundary([volumes[top_index]], recursive=False, oriented=False)
         for entity in top_surfaces:
             com = gmsh.model.occ.getCenterOfMass(entity[0], entity[1])
             if np.allclose(com, [(x1 - x0) / 2, y1, (z2 - z1) / 2 + z1]):
@@ -130,22 +131,20 @@ def generate_hex_boxes():
 
         for volume in volume_entities.keys():
             gmsh.model.addPhysicalGroup(
-                volume_entities[volume][0][0], [volume_entities[volume][0][1]],
-                tag=volume_entities[volume][1])
-            gmsh.model.setPhysicalName(volume_entities[volume][0][0],
-                                       volume_entities[volume][1], volume)
+                volume_entities[volume][0][0],
+                [volume_entities[volume][0][1]],
+                tag=volume_entities[volume][1],
+            )
+            gmsh.model.setPhysicalName(volume_entities[volume][0][0], volume_entities[volume][1], volume)
 
         for box in entities.keys():
             for surface in entities[box].keys():
-                gmsh.model.addPhysicalGroup(2, entities[box][surface][0],
-                                            tag=entities[box][surface][1])
-                gmsh.model.setPhysicalName(2, entities[box][surface][1], box
-                                           + ":" + surface)
+                gmsh.model.addPhysicalGroup(2, entities[box][surface][0], tag=entities[box][surface][1])
+                gmsh.model.setPhysicalName(2, entities[box][surface][1], box + ":" + surface)
         # Set mesh sizes on the points from the surface we are extruding
         bottom_nodes = gmsh.model.getBoundary([(2, bottom)], recursive=True, oriented=False)
         gmsh.model.occ.mesh.setSize(bottom_nodes, res)
-        top_nodes = gmsh.model.getBoundary([(2, top)],
-                                           recursive=True, oriented=False)
+        top_nodes = gmsh.model.getBoundary([(2, top)], recursive=True, oriented=False)
         gmsh.model.occ.mesh.setSize(top_nodes, 2 * res)
         # NOTE: Need to synchronize after setting mesh sizes
         gmsh.model.occ.synchronize()
@@ -174,14 +173,15 @@ def test_cube_contact(generate_hex_boxes, nonslip, get_assemblers):  # noqa: F81
     mesh, mt = mesh_data
     fdim = mesh.topology.dim - 1
     # Create functionspaces
-    V = fem.VectorFunctionSpace(mesh, ("Lagrange", 1))
+    V = fem.functionspace(mesh, ("Lagrange", 1, (mesh.geometry.dim,)))
 
     # Helper for orienting traction
 
     # Bottom boundary is fixed in all directions
     u_bc = fem.Function(V)
-    with u_bc.vector.localForm() as u_local:
+    with u_bc.x.petsc_vec.localForm() as u_local:
         u_local.set(0.0)
+    u_bc.x.petsc_vec.destroy()
 
     bottom_dofs = fem.locate_dofs_topological(V, fdim, mt.find(5))
     bc_bottom = fem.dirichletbc(u_bc, bottom_dofs)
@@ -189,8 +189,7 @@ def test_cube_contact(generate_hex_boxes, nonslip, get_assemblers):  # noqa: F81
     g_vec = [0, 0, -4.25e-1]
     if not nonslip:
         # Helper for orienting traction
-        r_matrix = dolfinx_mpc.utils.rotation_matrix(
-            [1 / np.sqrt(2), 1 / np.sqrt(2), 0], -theta)
+        r_matrix = dolfinx_mpc.utils.rotation_matrix([1 / np.sqrt(2), 1 / np.sqrt(2), 0], -theta)
 
         # Top boundary has a given deformation normal to the interface
         g_vec = np.dot(r_matrix, [0, 0, -4.25e-1])
@@ -202,6 +201,7 @@ def test_cube_contact(generate_hex_boxes, nonslip, get_assemblers):  # noqa: F81
         values[1] = g_vec[1]
         values[2] = g_vec[2]
         return values
+
     u_top = fem.Function(V)
     u_top.interpolate(top_v)
 
@@ -211,21 +211,20 @@ def test_cube_contact(generate_hex_boxes, nonslip, get_assemblers):  # noqa: F81
     bcs = [bc_bottom, bc_top]
 
     # Elasticity parameters
-    E = PETSc.ScalarType(1.0e3)
+    E = 1.0e3
     nu = 0
-    mu = fem.Constant(mesh, E / (2.0 * (1.0 + nu)))
-    lmbda = fem.Constant(mesh, E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu)))
+    mu = fem.Constant(mesh, default_scalar_type(E / (2.0 * (1.0 + nu))))
+    lmbda = fem.Constant(mesh, default_scalar_type(E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu))))
 
     # Stress computation
     def sigma(v):
-        return (2.0 * mu * ufl.sym(ufl.grad(v))
-                + lmbda * ufl.tr(ufl.sym(ufl.grad(v))) * ufl.Identity(len(v)))
+        return 2.0 * mu * ufl.sym(ufl.grad(v)) + lmbda * ufl.tr(ufl.sym(ufl.grad(v))) * ufl.Identity(len(v))
 
     # Define variational problem
     u = ufl.TrialFunction(V)
     v = ufl.TestFunction(V)
     a = ufl.inner(sigma(u), ufl.grad(v)) * ufl.dx
-    rhs = ufl.inner(fem.Constant(mesh, PETSc.ScalarType((0, 0, 0))), v) * ufl.dx
+    rhs = ufl.inner(fem.Constant(mesh, default_scalar_type((0, 0, 0))), v) * ufl.dx
     bilinear_form = fem.form(a)
 
     linear_form = fem.form(rhs)
@@ -240,11 +239,11 @@ def test_cube_contact(generate_hex_boxes, nonslip, get_assemblers):  # noqa: F81
     mpc = dolfinx_mpc.MultiPointConstraint(V)
     if nonslip:
         with Timer("~Contact: Create non-elastic constraint"):
-            mpc.create_contact_inelastic_condition(mt, 4, 9)
+            mpc.create_contact_inelastic_condition(mt, 4, 9, eps2=500 * np.finfo(default_scalar_type).resolution)
     else:
         with Timer("~Contact: Create contact constraint"):
             nh = dolfinx_mpc.utils.create_normal_approximation(V, mt, 4)
-            mpc.create_contact_slip_condition(mt, 4, 9, nh)
+            mpc.create_contact_slip_condition(mt, 4, 9, nh, eps2=500 * np.finfo(default_scalar_type).resolution)
 
     mpc.finalize()
 
@@ -259,23 +258,14 @@ def test_cube_contact(generate_hex_boxes, nonslip, get_assemblers):  # noqa: F81
 
     with Timer("~MPC: Solve"):
         solver.setOperators(A)
-        uh = b.copy()
-        uh.set(0)
-        solver.solve(b, uh)
+        uh = fem.Function(mpc.function_space)
+        uh.x.array[:] = 0
+        u_vec = uh.x.petsc_vec
+        solver.solve(b, u_vec)
 
-    uh.ghostUpdate(addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD)
+    u_vec.ghostUpdate(addv=PETSc.InsertMode.INSERT, mode=PETSc.ScatterMode.FORWARD)
     mpc.backsubstitution(uh)
 
-    # Write solution to file
-    # u_h = fem.Function(mpc.function_space)
-    # u_h.vector.setArray(uh.array)
-    # u_h.x.scatter_forward()
-    # u_h.name = "u_{0:.2f}".format(theta)
-    # import dolfinx.io as io
-    # with io.XDMFFile(comm, "output/rotated_cube3D.xdmf", "w") as outfile:
-    #     outfile.write_mesh(mesh)
-    #     outfile.write_function(u_h, 0.0, f"Xdmf/Domain/Grid[@Name='{mesh.name}'][1]")
-
     # Solve the MPC problem using a global transformation matrix
     # and numpy solvers to get reference values
     dolfinx_mpc.utils.log_info("Solving reference problem with global matrix (using numpy)")
@@ -296,7 +286,7 @@ def test_cube_contact(generate_hex_boxes, nonslip, get_assemblers):  # noqa: F81
         A_csr = dolfinx_mpc.utils.gather_PETScMatrix(A_org, root=root)
         K = dolfinx_mpc.utils.gather_transformation_matrix(mpc, root=root)
         L_np = dolfinx_mpc.utils.gather_PETScVector(L_org, root=root)
-        u_mpc = dolfinx_mpc.utils.gather_PETScVector(uh, root=root)
+        u_mpc = dolfinx_mpc.utils.gather_PETScVector(u_vec, root=root)
 
         if MPI.COMM_WORLD.rank == root:
             KTAK = K.T * A_csr * K
@@ -305,6 +295,10 @@ def test_cube_contact(generate_hex_boxes, nonslip, get_assemblers):  # noqa: F81
             d = scipy.sparse.linalg.spsolve(KTAK, reduced_L)
             # Back substitution to full solution vector
             uh_numpy = K @ d
-            assert np.allclose(uh_numpy, u_mpc)
+            atol = 1000 * np.finfo(default_scalar_type).resolution
+            nt.assert_allclose(uh_numpy, u_mpc, atol=atol)
+    L_org.destroy()
+    b.destroy()
+    solver.destroy()
 
     list_timings(comm, [TimingType.wall])
